/*
 * usb-vhci-hcd.h -- VHCI USB host controller driver header.
 *
 * Copyright (C) 2007-2008 Conemis AG Karlsruhe Germany
 * Copyright (C) 2007-2010 Michael Singer <michael@a-singer.de>
 * Copyright (C) 2016 Eltima Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef _EVEUSB_VHCI_HCD_H
#define _EVEUSB_VHCI_HCD_H

#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/timer.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <linux/platform_device.h>
#include <linux/usb.h>
#include <linux/usb/hcd.h>
#include <linux/device.h>
#include <asm/atomic.h>

// this is undefined in linux >= 2.6.35
#ifndef USB_PORT_FEAT_HIGHSPEED
#	define USB_PORT_FEAT_HIGHSPEED (10)
#endif

#include "eveusb-vhci.h"
#include "conf/eveusb-vhci.config.h"

struct eveusb_vhci_port
{
	u16 port_status;
	u16 port_change;
	u8 port_flags;
};

enum eveusb_vhci_rh_state
{
	USB_VHCI_RH_RESET     = 0,
	USB_VHCI_RH_SUSPENDED = 1,
	USB_VHCI_RH_RUNNING   = 2
} __attribute__((packed));

struct eveusb_vhci_device;

struct eveusb_vhci_ifc
{
	const char *ifc_desc;
	struct module *owner;
	size_t ifc_priv_size;

	// callbacks for backend drivers
	int (*init)(void *context, void *ifc_priv);
	void (*destroy)(void *ifc_priv);
	void (*wakeup)(struct eveusb_vhci_device *vdev);
};

struct eveusb_vhci_device
{
	const struct eveusb_vhci_ifc *ifc;
	struct platform_device *pdev;
	struct eveusb_vhci_hcd *vhc;

	u8 port_count;

	// private data for backend drivers
	unsigned long ifc_priv[0] __attribute__((aligned(sizeof(unsigned long))));
};

struct eveusb_vhci_urb_priv
{
	struct urb *urb;
	struct list_head urbp_list;
	atomic_t status;
};

struct eveusb_vhci_hcd
{
	struct eveusb_vhci_port *ports;
	u32 port_update;

	spinlock_t lock;

	atomic_t frame_num;
	enum eveusb_vhci_rh_state rh_state;

	// TODO: implement timer for incrementing frame_num every millisecond
	//struct timer_list timer;

	// urbs which are waiting to get fetched by user space are in this list
	struct list_head urbp_list_inbox;

	// urbs which were fetched by user space but not already given back are in this list
	struct list_head urbp_list_fetched;

	// urbs which were fetched by user space and not already given back, and which should be
	// canceled are in this list
	struct list_head urbp_list_cancel;

	// urbs which were fetched by user space and not already given back, and for which the
	// user space already knows about the cancelation state are in this list
	struct list_head urbp_list_canceling;

	u8 port_count;
};

static inline struct eveusb_vhci_device *pdev_to_vhcidev(struct platform_device *pdev)
{
	return pdev->dev.platform_data;
}

static inline void *vhcidev_to_ifc(struct eveusb_vhci_device *vdev)
{
	return &vdev->ifc_priv;
}

static inline struct eveusb_vhci_device *ifc_to_vhcidev(void *ifc)
{
	return container_of(ifc, struct eveusb_vhci_device, ifc_priv);
}

static inline struct eveusb_vhci_hcd *vhcidev_to_vhcihcd(struct eveusb_vhci_device *vdev)
{
	return vdev->vhc;
}

static inline struct platform_device *vhcidev_to_pdev(struct eveusb_vhci_device *vdev)
{
	return vdev->pdev;
}

static inline struct usb_hcd *vhcihcd_to_usbhcd(struct eveusb_vhci_hcd *vhc)
{
	return container_of((void *)vhc, struct usb_hcd, hcd_priv);
}

static inline struct usb_hcd *vhcidev_to_usbhcd(struct eveusb_vhci_device *vdev)
{
	return vhcihcd_to_usbhcd(vhcidev_to_vhcihcd(vdev));
}

static inline struct eveusb_vhci_hcd *usbhcd_to_vhcihcd(struct usb_hcd *hcd)
{
	return (struct eveusb_vhci_hcd *)&hcd->hcd_priv;
}

static inline struct device *usbhcd_to_dev(struct usb_hcd *hcd)
{
	return hcd->self.controller;
}

static inline struct device *vhcihcd_to_dev(struct eveusb_vhci_hcd *vhc)
{
	return usbhcd_to_dev(vhcihcd_to_usbhcd(vhc));
}

static inline struct device *vhcidev_to_dev(struct eveusb_vhci_device *vdev)
{
	return usbhcd_to_dev(vhcidev_to_usbhcd(vdev));
}

static inline struct platform_device *vhcihcd_to_pdev(struct eveusb_vhci_hcd *vhc)
{
	return to_platform_device(vhcihcd_to_dev(vhc));
}

static inline struct eveusb_vhci_device *vhcihcd_to_vhcidev(struct eveusb_vhci_hcd *vhc)
{
	return pdev_to_vhcidev(vhcihcd_to_pdev(vhc));
}

static inline struct eveusb_vhci_hcd *pdev_to_vhcihcd(struct platform_device *pdev)
{
	return vhcidev_to_vhcihcd(pdev_to_vhcidev(pdev));
}

static inline struct usb_hcd *pdev_to_usbhcd(struct platform_device *pdev)
{
	return vhcidev_to_usbhcd(pdev_to_vhcidev(pdev));
}

const char *eveusb_vhci_dev_name(struct eveusb_vhci_device *vdev);
int eveusb_vhci_dev_id(struct eveusb_vhci_device *vdev);
int eveusb_vhci_dev_busnum(struct eveusb_vhci_device *vdev);
void eveusb_vhci_maybe_set_status(struct eveusb_vhci_urb_priv *urbp, int status);
void eveusb_vhci_urb_giveback(struct eveusb_vhci_hcd *vhc, struct eveusb_vhci_urb_priv *urbp);
int eveusb_vhci_hcd_register(const struct eveusb_vhci_ifc *ifc, void *context, u8 port_count, struct eveusb_vhci_device **vdev_ret);
int eveusb_vhci_hcd_unregister(struct eveusb_vhci_device *vdev);
int eveusb_vhci_hcd_has_work(struct eveusb_vhci_hcd *vhc);
int eveusb_vhci_apply_port_stat(struct eveusb_vhci_hcd *vhc, u16 status, u16 change, u8 index);

int vhci_iocifc_init(void);
void vhci_iocifc_exit(void);

#endif
